#define UNWIND_HISTORY_TABLE_NONE   0
#define UNWIND_HISTORY_TABLE_GLOBAL 1
#define UNWIND_HISTORY_TABLE_LOCAL  2

//
// RaiseStatusIndex, Unwind, Exception are new to Vista.
//
// RaiseStatusIndex is set to the index into RtlpFunctionAddressTable for the
// raise exception routine for the processor mode.  In the user mode version
// of RtlpUnwindHistoryTable, this is RtlRaiseException.  In the kernel mode
// version of RtlpUnwindHistoryTable, this is RtlRaiseStatus.
//
// Note that RaiseStatusIndex is never used in a user-supplied unwind history
// table; it is only used by the global RtlpFunctionAddressTable that lives in
// ntoskrnl or ntdll.
//
// Unwind and Exception are hints to the function lookup logic.  These are
// typically supplied with a user-supplied unwind history table for use with a
// one-shot lookup operation (i.e. where RtlLookupFunctionEntry is only going
// to be called once, and not in a loop).  This may initially seem strange, as
// the documented purpose of the UNWIND_HISTORY_TABLE argument is as a cache
// that is used to speed up repeated address lookups for exception or unwind
// dispatching.  However, these two flags enable undocumented behaviors on
// Windows Vista (or later).
//
// Setting Unwind to TRUE causes an initial lookup against a flat array of
// function addresses stored in RtlpFunctionAddressTable (part of ntoskrnl or
// ntdll).
//

typedef struct _UNWIND_HISTORY_TABLE {
        ULONG                      Count;
        UCHAR                      Search;
        UCHAR                      RaiseStatusIndex;
        BOOLEAN                    Unwind;
        BOOLEAN                    Exception;
        ULONG64                    LowAddress;
        ULONG64                    HighAddress;
        UNWIND_HISTORY_TABLE_ENTRY Entry[ UNWIND_HISTORY_TABLE_SIZE ];
} UNWIND_HISTORY_TABLE, *PUNWIND_HISTORY_TABLE;

extern "C"
ULONG
NTAPI
RtlpExecuteHandlerForUnwind(
	__in    PEXCEPTION_RECORD   ExceptionRecord,
	__in    ULONG64             EstablisherFrame,
	__inout PCONTEXT            Context,
	__inout PDISPATCHER_CONTEXT DispatcherContext
	);

VOID
NTAPI
RtlUnwindEx(
	__in_opt ULONG64               TargetFrame,
	__in_opt ULONG64               TargetIp,
	__in_opt PEXCEPTION_RECORD     ExceptionRecord,
	__in     PVOID                 ReturnValue,
	__out    PCONTEXT              OriginalContext,
	__in_opt PUNWIND_HISTORY_TABLE HistoryTable
	)
{
	CONTEXT            UnwindContext;
	DISPATCHER_CONTEXT DispatcherContext;
	EXCEPTION_RECORD   DefaultExceptionRecord;
	ULONG64            ImageBase;
	ULONG64            EstablisherFrame;
	ULONG64            ControlPc;
	PRUNTIME_FUNCTION  FunctionEntry;
	PVOID              StackLimit, StackBase;
	PVOID              HandlerData;
	PEXCEPTION_ROUTINE UnwindHandler;
	PCONTEXT           ActiveUnwindContext;
	PCONTEXT           ActiveContext;
	PCONTEXT           Swap;
	ULONG              ExceptionFlags;
	ULONG              ScopeIndex;

	ActiveUnwindContext = &UnwindContext;
	ActiveContext       = OriginalContext;

	RtlpGetStackLimits(
		&StackLimit,
		&StackBase);
	RtlCaptureContext(
		OriginalContext);

	//
	// If we have a history table, mark it as in use for unwinding.
	//

	if (HistoryTable)
	{
		HistoryTable->Unwind = TRUE;
	}

	//
	// If we don't have an exception record, then we'll construct a default one
	// on the stack and use it instead.
	//

	if (!ExceptionRecord)
	{
		ExceptionRecord = &DefaultExceptionRecord;

		DefaultExceptionRecord.ExceptionCode    = STATUS_UNWIND;
		DefaultExceptionRecord.ExceptionAddress = (PVOID)OriginalContext->Rip;
		DefaultExceptionRecord.ExceptionRecord  = 0;
		DefaultExceptionRecord.NumberParameters = 0;
	}

	ExceptionFlags = EXCEPTION_UNWINDING;

	//
	// If we don't have a target frame, then mark this as an exit unwind.
	//

	if (!TargetFrame)
	{
		ExceptionFlags |= EXCEPTION_EXIT_UNWIND;
	}

	for (;;)
	{
		//
		// Look up the function entry for the current RIP value.
		//

		ControlPc = ActiveContext->Rip;

		FunctionEntry = RtlLookupFunctionEntry(
			ControlPc,
			&ImageBase,
			HistoryTable);

		if (FunctionEntry)
		{
			//
			// If we found a function entry, duplicate the context and perform
			// a virtual unwind.
			//

			RtlpCopyContext(
				ActiveUnwindContext,
				ActiveContext);

			UnwindHandler = RtlVirtualUnwind(
				UNW_FLAG_UHANDLER,
				ImageBase,
				ControlPc,
				FunctionEntry,
				ActiveUnwindContext,
				&HandlerData,
				&EstablisherFrame,
				0);

			if (EstablisherFrame & 7)
			{
				RtlRaiseStatus(STATUS_BAD_STACK);
			}

			if (EstablisherFrame <  (ULONG64)StackLimit ||
				EstablisherFrame >= (ULONG64)StackBase)
			{
#ifdef NTOS_KERNEL_RUNTIME
				//
				// If we're in the context of a DPC, check the DPC stack.
				//

				if (KeGetPcr()->Prcb->DpcRoutineActive)
				{
					StackLimit = *(PULONG64)((PUCHAR)StackBase - 0x28);
					StackBase  = *(PULONG64)((PUCHAR)StackBase - 0x20);

					if (!StackBase                              ||
						EstablisherFrame <  (ULONG64)StackLimit ||
						EstablisherFrame >= (ULONG64)StackBase)
					{
						RtlRaiseStatus(STATUS_BAD_STACK);
					}
				}
#else
				RtlRaiseStatus(STATUS_BAD_STACK);
#endif
			}

			//
			// If we're a target unwind, then make that the target frame is
			// within the call stack of the current function.
			//

			if (TargetFrame)
			{
				if (TargetFrame < EstablisherFrame)
				{
					RtlRaiseStatus(STATUS_BAD_STACK);
				}
			}

			//
			// If we have an unwind handler to call then do so now.
			//

			if (UnwindHandler)
			{
				DispatcherContext.TargetIp = TargetIp;

				ScopeIndex = 0;

				do
				{
					//
					// If we've reached the target frame, then indicate to the
					// unwind handler that this is a target unwind.
					//

					if (TargetFrame == EstablisherFrame)
					{
						ExceptionFlags |= EXCEPTION_TARGET_UNWIND;
					}

					//
					// Set up for the call to the unwind handler.
					//

					ExceptionRecord->ExceptionFlags = ExceptionFlags;

					ExceptionFlags &= ~(EXCEPTION_TARGET_UNWIND | EXCEPTION_COLLIDED_UNWIND);

					ActiveContext->Rax = (ULONG64)ReturnValue;

					DispatcherContext.ImageBase        = ImageBase;
					DispatcherContext.HandlerData      = HandlerData;
					DispatcherContext.HistoryTable     = HistoryTable;
					DispatcherContext.ControlPc        = ControlPc;
					DispatcherContext.FunctionEntry    = FunctionEntry;
					DispatcherContext.EstablisherFrame = EstablisherFrame;
					DispatcherContext.ContextRecord    = ActiveContext;
					DispatcherContext.LanguageHandler  = UnwindHandler;
					DispatcherContext.ScopeIndex       = ScopeIndex;

					switch (RtlpExecuteHandlerForUnwind(
						ExceptionRecord,
						EstablisherFrame,
						ActiveContext,
						&DispatcherContext))
					{

					case ExceptionContinueSearch:
						//
						// Swap the active context with the unwound context if
						// we haven't yet reached our target frame.
						//

						if (EstablisherFrame != TargetFrame)
						{
							Swap                = ActiveContext;
							ActiveContext       = ActiveUnwindContext;
							ActiveUnwindContext = Swap;
						}
						break;

					case ExceptionCollidedUnwind:
						//
						// If we caught an exception during unwind (collided
						// unwind), then fetch the information from the newly
						// updated DISPATCHER_CONTEXT, and reset our context
						// pointers.
						//

						ImageBase           = DispatcherContext.ImageBase;
						ControlPc           = DispatcherContext.ControlPc;
						FunctionEntry       = DispatcherContext.FunctionEntry;

						ActiveUnwindContext = &UnwindContext;
						ActiveContext       = OriginalContext;

						RtlpCopyContext(
							ActiveContext,
							DispatcherContext.ContextRecord);
						RtlpCopyContext(
							ActiveUnwindContext,
							ActiveContext);

						//
						// Execute a virtual unwind to locate the next call
						// frame in the call stack.
						//

						RtlVirtualUnwind(
							UNW_FLAG_NHANDLER,
							ImageBase,
							ControlPc,
							FunctionEntry,
							ActiveUnwindContext,
							&HandlerData,
							&EstablisherFrame,
							0);

						//
						// Most of the data we're interested in given by the
						// DispatcherContext of the unwind instance that
						// participated in the collided unwind.  As such,
						// update our internal state accordingly.
						//

						HandlerData       = DispatcherContext.HandlerData;
						ScopeIndex        = DispatcherContext.ScopeIndex;
						UnwindHandler     = DispatcherContext.LanguageHandler;
						HistoryTable      = DispatcherContext.HistoryTable;
						EstablisherFrame  = DispatcherContext.EstablisherFrame;
						ExceptionFlags   |= EXCEPTION_COLLIDED_UNWIND;
						break;

					default:
						RtlRaiseStatus(STATUS_INVALID_DISPOSITION);
						break;

					}


				} while (ExceptionFlags & EXCEPTION_COLLIDED_UNWIND) ;
			}
			else
			{
				//
				// Swap the active context with the unwound context if
				// we haven't yet reached our target frame.
				//

				if (EstablisherFrame != TargetFrame)
				{
					Swap                = ActiveContext;
					ActiveContext       = ActiveUnwindContext;
					ActiveUnwindContext = Swap;
				}
			}
		}
		else
		{
			//
			// If we don't have a valid RUNTIME_FUNCTION, then this must be a
			// leaf function.  Get the return address from [Rsp], and adjust
			// our context accordingly.
			//

			ActiveContext->Rip  = *(PULONG64)(ActiveContext->Rsp);
			ActiveContext->Rsp += sizeof(ULONG64);
		}


		if (!(EstablisherFrame & 7))
		{
#ifdef NTOS_KERNEL_RUNTIME
			//
			// If we're a DPC, check to make sure that we are within a DPC stack
			//

			if (EstablisherFrame < (ULONG64)StackLimit ||
				EstablisherFrame >= (ULONG64)StackBase)
			{
				if (KeGetPcr()->Prcb->DpcRoutineActive)
				{
					StackLimit = *(PULONG64)((PUCHAR)StackBase - 0x28);
					StackBase  = *(PULONG64)((PUCHAR)StackBase - 0x20);

					if (StackBase                               &&
						EstablisherFrame >= (ULONG64)StackLimit &&
						EstablisherFrame <  (ULONG64)StackBase)
					{
						if (EstablisherFrame != TargetFrame)
							continue;

						break;
					}
				}
			}
			else
			{
				if (EstablisherFrame != TargetFrame)
					continue;

				break;
			}
#else
			if (EstablisherFrame != TargetFrame)
				continue;

			break;
#endif
		}

		//
		// If we haven't reached our target frame by now, then something is
		// wrong.  Raise the appropriate exception if this is the case.
		//
		// This is typically a nonrecoverable error.
		//

		if (EstablisherFrame != TargetFrame)
		{
			if (ControlPc != ActiveContext->Rip)
			{
				ZwRaiseException(
					ExceptionRecord,
					ActiveContext,
					FALSE);
			}
			else
			{
				RtlRaiseStatus(STATUS_BAD_FUNCTION_TABLE);
			}
		}

		break;
	}

	//
	// We're finished.  Activate the new context.
	//

	ActiveContext->Rax = (ULONG64)ReturnValue;

	//
	// If an exception handler didn't mark the operation as an unwind
	// consolidate, then set execution to resume at TargetIp.  Otherwise, we
	// simply resume at the old Rip value.
	//

	if (ExceptionRecord->ExceptionCode != STATUS_UNWIND_CONSOLIDATE)
	{
		ActiveContext->Rip = TargetIp;
	}

	RtlRestoreContext(
		ActiveContext,
		ExceptionRecord);
}
